#include <u.h>
#include <libc.h>
#include <stdio.h>
#include <thread.h>
#include "dat.h"
#include "fns.h"

enum{
	Nbuf = 8
};
static int afd, wpos, stid = -1;
static Channel *schan;
static QLock sndlock;

static void
sproc(void *)
{
	int n;

	for(;;){
		if(recv(schan, nil) < 0)
			break;
		if((n = write(afd, dma.buffer, SNBUF)) < 0)
			break;
		qlock(&sndlock);
		wpos += n;
		qunlock(&sndlock);
	}
}

qboolean
SNDDMA_Init(void)
{
	if(stid >= 0 || COM_CheckParm("-nosound"))
		return false;
	if((afd = open("/dev/audio", OWRITE)) < 0){
		fprint(2, "SNDDMA_Init: %r\n");
		return false;
	}
	dma.samplebits = SAMPLESZ;
	dma.speed = RATE;
	dma.channels = 2;
	dma.samples = NSAMPLE;
	dma.submission_chunk = 1;
	if((dma.buffer = mallocz(SNBUF, 1)) == nil)
		sysfatal("mallocz %ud: %r\n", SNBUF);
	dma.samplepos = 0;
	wpos = 0;

	schan = chancreate(sizeof(int), Nbuf);
	if((stid = proccreate(sproc, nil, 8192)) < 0){
		SNDDMA_Shutdown();
		sysfatal("proccreate sproc: %r\n");
	}
	return true;
}

int
SNDDMA_GetDMAPos(void)
{
	if(stid < 0)
		return 0;
	qlock(&sndlock);
	dma.samplepos = wpos / SAMPLEB;
	qunlock(&sndlock);
	return dma.samplepos;
}

void
SNDDMA_Shutdown(void)
{
	if(stid < 0)
		return;

	threadint(stid);
	close(afd);
	if(schan != nil){
		chanfree(schan);
		schan = nil;
	}
	free(dma.buffer);
	stid = -1;
}

void
SNDDMA_Submit(void)
{
	if(nbsend(schan, nil) < 0){
		fprint(2, "SNDDMA_Submit: %r\n");
		SNDDMA_Shutdown();
	}
}

void
SNDDMA_BeginPainting(void)
{
}
